import { providers } from "ethers";
export interface ErrorResponse {
    code?: string;
    msg: string;
}
export enum ChainTypes {
    ETH = "ethereum",
    SOL = "solana",
    FLOW = "flow"
}
export type ChainType = `${ChainTypes}`;
export type ChainInfo = {
    chainSlug: string;
    chainId: number;
    chainType: ChainType;
    chainName: string;
    publicRPC: string;
    authorizationSupported: boolean;
};
export enum SigningMessageFormat {
    SIMPLE = "simple",
    SIWE = "siwe"
}
export interface NonceRequest {
    chain: string;
    walletAddress: string;
    locale?: string;
}
export interface NonceResponse {
    nonce: string;
    statement: string;
    format: `${SigningMessageFormat}`;
}
export interface AuthRequirements {
    contractAddress?: string;
    minTokenBalance?: number | string;
    allowedWallets?: string[];
    tokenIds?: string[];
    collection?: string | string[];
    creatorAddress?: string;
}
export type TokenRequirementsBalances = {
    contractAddress?: Record<string, string>;
    collection?: Record<string, string>;
    tokenIds?: Record<string, string>;
    creatorAddress?: Record<string, string>;
};
export interface AuthenticatedUser {
    chain: string;
    walletAddress: string;
    displayAddress: string;
    contractAddress?: string;
    tokenBalances?: TokenRequirementsBalances;
}
export interface AuthState {
    accessToken: string;
    user: AuthenticatedUser;
}
export interface AccessTokenPayload extends AuthenticatedUser {
    iat: number;
    exp: number;
    iss: string;
    sub: string;
    aud: string;
    tid: string;
}
export interface SigningMessageContext {
    domain: string;
    uri: string;
    chainId: number;
    issuedAt: string;
    chainType: ChainType;
    locale?: string;
}
export interface SigningMessageRequestSimple extends NonceResponse {
    walletAddress: string;
}
export interface SigningMessageRequestSIWE extends SigningMessageRequestSimple, SigningMessageContext {
}
export type SigningMessageRequest = SigningMessageRequestSimple | SigningMessageRequestSIWE;
export interface ConnectRequest {
    chain?: string;
    doAuth?: boolean;
    requirements?: AuthRequirements;
}
export type ConnectProvider = providers.ExternalProvider | providers.JsonRpcFetchFunc | any;
export interface ConnectResponse {
    walletAddress: string;
    signature: string;
    provider: ConnectProvider;
    context?: SigningMessageContext;
    chain: string;
    auth?: AuthState;
}
export interface AppState extends Record<string, any> {
    returnTo?: string;
}
export interface LoginRequest extends AuthRequirements {
    chain?: string;
    walletAddress?: string;
    signature?: string;
    context?: SigningMessageContext;
}
export interface LoginOptions {
    redirectURI?: string;
    appState?: AppState;
}
export interface LoginCallbackResponse extends AuthState {
    appState: AppState;
}
export type AuthorizationResponseMode = "code" | "web_message";
export interface AuthorizationURLRequest extends AuthRequirements {
    redirectURI: string;
    state: string;
    codeChallenge: string;
    chain?: string;
    walletAddress?: string;
    signature?: string;
    responseMode: AuthorizationResponseMode;
    locale?: string;
    context?: SigningMessageContext;
}
export interface AuthorizationServerWebResponse {
    state: string;
    code?: string;
    error?: string;
    error_description?: string;
}
export interface AuthRequest {
    chain: string;
    walletAddress: string;
    signature: string;
    requirements?: AuthRequirements;
    context?: SigningMessageContext;
}
export const randomState: () => string;
export const hasAuthorizationCodeParams: (url: string) => boolean;
export interface AuthorizationCodeParams {
    state: string | null;
    code: string | null;
    error: string | null;
    error_description: string | null;
}
export const parseAuthorizationCodeParams: (url: string) => AuthorizationCodeParams;
export const defaultLoginRedirectCallback: (appState?: AppState) => void;
export type PicketTheme = "light" | "dark" | "auto";
export interface PicketOptions {
    theme?: PicketTheme;
    baseURL?: string;
}
export const DEFAULT_THEME = "light";
export const API_VERSION = "v1";
export class Picket {
    #private;
    baseURL: string;
    theme: PicketTheme;
    constructor(apiKey: string, { baseURL, theme }?: PicketOptions);
    /**
     * nonce
     * Function for retrieving nonce for a given user
     */
    nonce({ walletAddress, chain, locale, }: NonceRequest): Promise<NonceResponse>;
    /**
     * auth
     * Function for initiating auth / token gating
     */
    auth({ chain, walletAddress, signature, requirements, context, }: AuthRequest): Promise<AuthState>;
    /**
     * authz
     * Function for checking if a given user is authorized (aka meets the requirements)
     */
    authz({ accessToken, requirements, revalidate, }: {
        accessToken: string;
        requirements: AuthRequirements;
        revalidate?: boolean;
    }): Promise<AuthState>;
    /**
     * Validate
     * Validate the given access token and requirements
     */
    validate(accessToken: string, requirements?: AuthRequirements): Promise<AccessTokenPayload>;
    /**
     * chainInfo
     * Function for retrieving chain information
     */
    chainInfo(chain: string): Promise<ChainInfo>;
    /**
     * chains
     * Function for retrieving information on supported chains
     */
    chains(): Promise<ChainInfo[]>;
    /**
     * login
     * Login with your wallet, and optionally, specify login requirements
     */
    login(req?: LoginRequest): Promise<AuthState>;
    /**
     * loginWithTrustedWalletProivder
     * Login with a trusted wallet provider that supports Sign-In With Ethereum (EIP-4361) or similar standard
     */
    loginWithTrustedWalletProivder({ chain, walletAddress, signature, context, ...requirements }?: LoginRequest): Promise<AuthState>;
    /**
     * getAuthorizationURL
     * getAuthorizationURL returns the authorization URL for the PKCE authorization parameters.
     */
    getAuthorizationURL({ chain, walletAddress, signature, contractAddress, tokenIds, collection, creatorAddress, minTokenBalance, allowedWallets, redirectURI, codeChallenge, state, responseMode, locale, context, }: AuthorizationURLRequest): string;
    /**
     * loginWithRedirect
     * loginWithRedirect starts the OAuth2.0 PKCE flow
     */
    loginWithRedirect({ chain, walletAddress, signature, tokenIds, contractAddress, minTokenBalance, allowedWallets, context, }?: LoginRequest, { redirectURI, appState }?: LoginOptions): Promise<void>;
    /**
     * oauth2AuthorizationCodeToken
     * oauth2AuthorizationCodeToken implements the final step of the OAuth2.0 PKCE flow and exchanges an authorization code for an access token.
     */
    oauth2AuthorizationCodeToken({ code, codeVerifier, redirectURI, }: {
        code: string;
        codeVerifier: string;
        redirectURI: string;
    }): Promise<AuthState>;
    /**
     * handleLoginRedirect
     * handleLoginRedirect fetches an access token after a successful authorization redirect.
     * If there are no authorization code query parameters or there are error query parameters, it will raise an error.
     */
    handleLoginRedirect(url?: string): Promise<LoginCallbackResponse>;
    /**
     * loginWithPopup
     * loginWithPopup starts the OAuth2.0 PKCE flow
     *
     * Implements https://datatracker.ietf.org/doc/html/draft-sakimura-oauth-wmrm-00#section-2.3
     */
    loginWithPopup({ chain, walletAddress, signature, contractAddress, tokenIds, minTokenBalance, allowedWallets, context, }?: LoginRequest, { redirectURI }?: LoginOptions): Promise<AuthState>;
    static createSigningMessage(args: SigningMessageRequest): string;
    /**
     * connect
     * Convenience function to connect wallet and sign nonce, prompts user to connect wallet and returns wallet object
     */
    connect({ chain, doAuth, requirements, }: ConnectRequest): Promise<ConnectResponse>;
    /**
     * logout
     * Clears authentication information
     */
    logout(): Promise<void>;
    /**
     * authState
     * get user auth information if it exists
     */
    authState(): Promise<AuthState | null>;
    /**
     * isCurrentUserAuthorized
     * Does the current user authorized given the requirements?
     */
    isCurrentUserAuthorized({ requirements, revalidate, }: {
        requirements: AuthRequirements;
        revalidate?: boolean;
    }): Promise<boolean>;
    /**
     * meetsAuthRequirements
     * Does the given user meet the authorization requirements?
     * Synchronous, helper function for users to check easily if the user meets the auth requirements.
     * Example usage is for setting "disabled" property on a button.
     */
    static meetsAuthRequirements({ user, requirements, }: {
        user: AuthenticatedUser;
        requirements: AuthRequirements;
    }): boolean;
}
export default Picket;

//# sourceMappingURL=index.d.ts.map
